from ..utils.streamlit_utils import st
import pandas as pd
import numpy as np
from scipy.stats import skew, boxcox
from imblearn.over_sampling import RandomOverSampler
from imblearn.under_sampling import RandomUnderSampler
from imblearn.combine import SMOTEENN
from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler

@st.cache_data
def log_transform(x):
    if np.any(x < 0):
        st.error(f"Negative or zero values encountered in '{x.name}' during Log transformation and hence cannot be applied")
        return x
    return np.log1p(x)

@st.cache_data
def boxcox_transform(x):
    if np.any(x <= 0):
        st.error(f"Negative or zero values encountered in '{x.name}' during Box-Cox transformation an hence cannot be applied")
        return x
    return boxcox(x - x.min() + 1)[0]

@st.cache_data
def sqrt_transform(x):
    if np.any(x < 0):
        st.error(f"Negative values encountered in '{x.name}' during Square Root transformation and hence transformation cannot be applied")
        return x
    return np.sqrt(x)

@st.cache_data
def reciprocal_transform(x):
    return 1 / (x + 1e-6)

transformations = {
    'Log Transformation': log_transform,
    'Box-Cox Transformation': boxcox_transform,
    'Square Root Transformation': sqrt_transform,
    'Reciprocal Transformation': reciprocal_transform
}

class DataPreprocessing:
    def __init__(_self, df, target):
        _self.df = df.copy()
        _self.target = target

    @st.cache_data
    def transform_columns(_self, method, S):
        skew_values = _self.df.apply(lambda x: skew(x))
        high_skew_columns = skew_values[skew_values > S].index

        transform = transformations[method]
        _self.df[high_skew_columns] = _self.df[high_skew_columns].apply(transform)
        
        return _self.df

    @st.cache_data
    def balance_dataset(_self, sample_tech, sample_st):
        X, y = _self.df.drop(_self.target, axis=1), _self.df[_self.target]
        if sample_tech == "Over Sampling":
            X_resampled, y_resampled = RandomOverSampler(random_state=0, sampling_strategy=sample_st).fit_resample(X, y)
        elif sample_tech == "Under Sampling":
            X_resampled, y_resampled = RandomUnderSampler(random_state=0, sampling_strategy=sample_st).fit_resample(X, y)
        elif sample_tech == "Combined":
            X_resampled, y_resampled = SMOTEENN(random_state=0, sampling_strategy=sample_st).fit_resample(X, y)
        _self.df = pd.DataFrame(X_resampled, columns=_self.df.columns[:-1])
        _self.df[_self.target] = y_resampled
        return _self.df

    @st.cache_data
    def scale_dataset(_self, scale_method):
        X, y = _self.df.drop(_self.target, axis = 1), _self.df[_self.target]
        if scale_method == "Standard Scaling":
            scaler = StandardScaler()
            X_scaled = scaler.fit_transform(X)
        elif scale_method == "Min-Max Scaling":
            scaler = MinMaxScaler()
            X_scaled = scaler.fit_transform(X)
        elif scale_method == "Robust Scaling":
            scaler = RobustScaler()
            X_scaled = scaler.fit_transform(X)
        _self.df = pd.DataFrame(X_scaled, columns=_self.df.columns[:-1])
        _self.df[_self.target] = y
        return _self.df


from ..utils.streamlit_utils import st
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from xgboost import XGBClassifier
from lightgbm import LGBMClassifier
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV

class HyperparameterTuning:
    def __init__(_self, X_train, y_train):
        _self.X_train = X_train
        _self.y_train = y_train

    @st.cache_data(show_spinner="Tuning Hyperparameters...")
    def tune_models(_self, models):
        if "Logistic Regression" in models:
            with st.spinner("Finding best parameters for Logistic Regression model..."):
                param_grid = {"penalty" : ['l1', 'l2', 'elasticnet'], 
                              "C" : [0.1, 1, 10], 
                              "solver" : ['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga']}
                grid = GridSearchCV(LogisticRegression(), param_grid, cv=5, scoring='accuracy')
                grid.fit(_self.X_train, _self.y_train)
                col1, col2 = st.columns(2)
                with col1:
                    st.write("Default Hyperparameters:")
                    st.dataframe(st.session_state.default_params['Logistic Regression'])
                with col2:
                    st.write("Best Hyperparameters Found:")
                    params_df = pd.DataFrame.from_dict(grid.best_params_, orient='index', columns=['Value'])
                    params_df.reset_index(inplace=True)
                    params_df.columns = ['Hyperparameter', 'Value']
                    st.dataframe(params_df)
            with st.spinner("Training Logistic Regression model..."):
                lr = grid.best_estimator_
                lr.fit(_self.X_train, _self.y_train)
                st.success("Logistic Regression model training complete!")
                st.session_state.tuned_models.append(lr)
                st.session_state.tuned_models_dict['Logistic Regression'] = lr

        if "Decision Tree" in models:
            with st.spinner("Finding best parameters for Decision Tree model..."):
                param_grid = {"max_depth": range(5, 31, 5),
                              "min_samples_leaf": [5, 7, 10],
                              "max_leaf_nodes": [5, 10, 15],
                              "min_impurity_decrease": [0.01, 0.1]}
                grid = GridSearchCV(DecisionTreeClassifier(), param_grid, cv=3, scoring='accuracy')
                grid.fit(_self.X_train, _self.y_train)
                col1, col2 = st.columns(2)
                with col1:
                    st.write("Default Hyperparameters:")
                    st.dataframe(st.session_state.default_params['Decision Tree'])
                with col2:
                    st.write("Best Hyperparameters Found:")
                    params_df = pd.DataFrame.from_dict(grid.best_params_, orient='index', columns=['Value'])
                    params_df.reset_index(inplace=True)
                    params_df.columns = ['Hyperparameter', 'Value']
                    st.dataframe(params_df)
            with st.spinner("Training Decision Tree model..."):
                dtree = grid.best_estimator_
                dtree.fit(_self.X_train, _self.y_train)
                st.success("Decision Tree model training complete!")
                st.session_state.tuned_models.append(dtree)
                st.session_state.tuned_models_dict['Decision Tree'] = dtree

        if "Random Forest" in models:
            with st.spinner("Finding best parameters for Random Forest model..."):
                param_grid = {
                    'n_estimators': [100, 300, 500],
                    'max_depth': [5, 10, 20, 40],
                    'min_samples_split': [2, 5, 10],
                    'min_samples_leaf': [1, 2, 4],
                }
                grid = RandomizedSearchCV(RandomForestClassifier(), param_grid, cv=3, scoring='accuracy')
                grid.fit(_self.X_train, _self.y_train)
                col1, col2 = st.columns(2)
                with col1:
                    st.write("Default Hyperparameters:")
                    st.dataframe(st.session_state.default_params['Random Forest'])
                with col2:
                    st.write("Best Hyperparameters Found:")
                    params_df = pd.DataFrame.from_dict(grid.best_params_, orient='index', columns=['Value'])
                    params_df.reset_index(inplace=True)
                    params_df.columns = ['Hyperparameter', 'Value']
                    st.dataframe(params_df)
            with st.spinner("Training Random Forest model..."):
                rf = grid.best_estimator_
                rf.fit(_self.X_train, _self.y_train)
                st.success("Random Forest model training complete!")
                st.session_state.tuned_models.append(rf)
                st.session_state.tuned_models_dict['Random Forest'] = rf

        if "XGBoost" in models:
            with st.spinner("Finding best parameters for XGBoost model..."):
                param_grid = {
                    'n_estimators': [50, 100, 200, 500],
                    'learning_rate': [0.01, 0.1, 0.5],
                    'max_depth': [3, 5, 7, 10],
                    'colsample_bytree': [0.5, 1],
                    'gamma': [0.0, 0.1, 0.2],
                    'min_child_weight': [1, 3, 5]
                }
                grid = GridSearchCV(XGBClassifier(), param_grid, cv=3, scoring='accuracy')
                grid.fit(_self.X_train, _self.y_train)
                col1, col2 = st.columns(2)
                with col1:
                    st.write("Default Hyperparameters:")
                    st.dataframe(st.session_state.default_params['XGBoost'])
                with col2:
                    st.write("Best Hyperparameters Found:")
                    params_df = pd.DataFrame.from_dict(grid.best_params_, orient='index', columns=['Value'])
                    params_df.reset_index(inplace=True)
                    params_df.columns = ['Hyperparameter', 'Value']
                    st.dataframe(params_df)
            with st.spinner("Training XGBoost model..."):
                xgb = grid.best_estimator_
                xgb.fit(_self.X_train, _self.y_train)
                st.success("XGBoost model training complete!")
                st.session_state.tuned_models.append(xgb)
                st.session_state.tuned_models_dict['XGBoost'] = xgb

        if "AdaBoost" in models:
            with st.spinner("Finding best parameters for AdaBoost model..."):
                param_grid = {
                    'n_estimators': [50, 100, 150, 200],
                    'learning_rate': [0.001, 0.01, 0.1, 1.0]
                }
                grid = GridSearchCV(AdaBoostClassifier(), param_grid, cv=3, scoring='accuracy')
                grid.fit(_self.X_train, _self.y_train)
                col1, col2 = st.columns(2)
                with col1:
                    st.write("Default Hyperparameters:")
                    st.dataframe(st.session_state.default_params['AdaBoost'])
                with col2:
                    st.write("Best Hyperparameters Found:")
                    params_df = pd.DataFrame.from_dict(grid.best_params_, orient='index', columns=['Value'])
                    params_df.reset_index(inplace=True)
                    params_df.columns = ['Hyperparameter', 'Value']
                    st.dataframe(params_df)
            with st.spinner("Training AdaBoost model..."):
                ada = grid.best_estimator_
                ada.fit(_self.X_train, _self.y_train)
                st.success("AdaBoost model training complete!")
                st.session_state.tuned_models.append(ada)
                st.session_state.tuned_models_dict['AdaBoost'] = ada

        if "LightGBM" in models:
            with st.spinner("Finding best parameters for LightGBM model..."):
                param_grid = {
                    'n_estimators': [50, 100, 200, 500],
                    'learning_rate': [0.01, 0.1, 0.2],
                    'max_depth': [3, 5, 7, 15],
                    'colsample_bytree': [0.3, 0.7],
                }
                grid = GridSearchCV(LGBMClassifier(), param_grid, cv=3, scoring='accuracy')
                grid.fit(_self.X_train, _self.y_train)
                col1, col2 = st.columns(2)
                with col1:
                    st.write("Default Hyperparameters:")
                    st.dataframe(st.session_state.default_params['LightGBM'])
                with col2:
                    st.write("Best Hyperparameters Found:")
                    params_df = pd.DataFrame.from_dict(grid.best_params_, orient='index', columns=['Value'])
                    params_df.reset_index(inplace=True)
                    params_df.columns = ['Hyperparameter', 'Value']
                    st.dataframe(params_df)
            with st.spinner("Training LightGBM model..."):
                lgbm = grid.best_estimator_
                lgbm.fit(_self.X_train, _self.y_train)
                st.success("LightGBM model training complete!")
                st.session_state.tuned_models.append(lgbm)
                st.session_state.tuned_models_dict['LightGBM'] = lgbm


from ..utils.streamlit_utils import st
import plotly.express as px

class DataVisualization:
    def __init__(self, df):
        self.df = df
        self.plot_colors = px.colors.sequential.YlOrRd[::-2]

    def visualize(self):
        if "target" in st.session_state:
            target = st.session_state.target
            # Correlation Analysis
            # st.subheader("Correlation Matrix")
            with st.popover("Correlation Matrix", use_container_width=True):
                fig = px.imshow(self.df.corr(numeric_only=True), width = 980, text_auto = ".2f", color_continuous_scale="electric")
                fig.update_layout(autosize=False, width=700, height=600, margin=dict(l=0, r=0, t=0, b=0))
                st.plotly_chart(fig)

            # Box Plots
            # st.subheader("Univariate Analysis (Box Plot)")
            with st.popover("Univariate Analysis (Box Plot)", use_container_width=True):
                all_cols_less_40 = [col for col in self.df.columns if self.df[col].nunique() < 40]
                cols_to_show = st.multiselect("Select columns to show", self.df.drop(target, axis = 1).columns, default=self.df.drop(target, axis = 1).columns[0], key = 'key1')
                for col in cols_to_show:
                    st.write(col)
                    fig = px.box(self.df, y=col, width=980, color_discrete_sequence=self.plot_colors[1:])
                    fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
                    st.plotly_chart(fig)

            # Bar Plots
            # st.subheader("Univariate Analysis (Bar Chart)")
            with st.popover("Univariate Analysis (Bar Chart)", use_container_width=True):
                all_cols_less_40 = [col for col in self.df.columns if self.df[col].nunique() < 40]
                cols_to_show = st.multiselect("Select columns to show", all_cols_less_40, default=all_cols_less_40[0], key = 'key2')
                for col in cols_to_show:
                    st.write(col)
                    st.bar_chart(self.df[col].value_counts())

            # Dist Plots Against Target Variable
            # st.subheader("Bivariate Analysis (Distribution Plots against Target)")
            with st.popover("Bivariate Analysis (Distribution Plots against Target)", use_container_width=True):
                x_col = st.selectbox("Select x column", self.df.drop(target, axis = 1).columns, index = 0, key = 'key3')
                fig = px.histogram(self.df, x=x_col, color=target, marginal="violin", hover_data=self.df.columns, opacity=0.5)
                st.plotly_chart(fig)

            # Scatter Plots
            all_cols_more_40 = [col for col in self.df.columns if self.df[col].nunique() >= 40]
            if len(all_cols_more_40) > 1:
                # st.subheader("Bivariate Analysis (Scatter plot between two columns)")
                with st.popover("Bivariate Analysis (Scatter plot between two columns)", use_container_width=True):
                    x_col = st.selectbox("Select x column", all_cols_more_40, index=0, key = 'key4')
                    y_col = st.selectbox("Select y column", all_cols_more_40, index=1, key = 'key5')
                    fig = px.scatter(self.df, x=x_col, y=y_col, width=980, color_discrete_sequence=self.plot_colors[1:])
                    fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
                    st.plotly_chart(fig)


from ..utils.streamlit_utils import st

class DataSummarization:
    def __init__(_self, df):
        _self.df = df

    def summarize(_self):
        if _self.df is not None:
            desc_stats = _self.df.describe().T
            additional_percentiles = [0.01, 0.05, 0.10, 0.90, 0.95, 0.99]
            for percentile in additional_percentiles:
                desc_stats[f'P{int(percentile * 100)}'] = _self.df.quantile(percentile)
            desc_stats['missing'] = _self.df.isna().sum()
            desc_stats = desc_stats.rename(columns = {'min': 'P0 (min)', '25%': 'P25', '50%': 'P50 (median)', '75%': 'P75', 'max': 'P100 (max)'})
            st.write(desc_stats[['count', 'missing', 'mean', 'std', 'P0 (min)', 'P1', 'P5', 'P10', 'P25', 'P50 (median)', 'P75', 'P90', 'P95', 'P99', 'P100 (max)']])
        
            st.write(f"Total number of missing values (entire row missing): :red[{_self.df.isna().sum().sum()}]")
            st.write(f"Total number of duplicated values (entire row duplicated): :red[{_self.df.duplicated().sum().sum()}]")


from ..utils.streamlit_utils import st
import os, io
import pandas as pd
import zipfile

class DataInput:
    def __init__(_self, uploaded_file, key_suffix = None):
        _self.uploaded_file = uploaded_file
        _self.key_suffix = key_suffix

    def upload_dataset(_self):
        st.write("**Filename:**", _self.uploaded_file.name)
        st.write("**Type:**", _self.uploaded_file.type)

        has_header = st.toggle('Does the dataset have a header?', value = True, key=f'header_toggle_{_self.key_suffix}')
        header_param = 0 if has_header else None

        file_extension = os.path.splitext(_self.uploaded_file.name)[1]

        if file_extension == '.csv':
            df = pd.read_csv(_self.uploaded_file, header = header_param)
        elif file_extension == '.tsv':
            df = pd.read_csv(_self.uploaded_file, sep = '\t', header = header_param)
        elif file_extension == '.txt':
            df = pd.read_csv(_self.uploaded_file, sep = ' ', header = header_param)
        elif file_extension == '.xlsx':
            df = pd.read_excel(_self.uploaded_file, header = header_param)
        elif file_extension == '.zip':
            with zipfile.ZipFile(_self.uploaded_file, 'r') as z:
                # Assuming there's only one file in the zip archive, and it's a csv file
                try:
                    with z.open(z.namelist()[0]) as f:
                        df = pd.read_csv(f, header = header_param)
                except:
                    st.error("Please upload a zip file containing a single CSV file")
    
        if not has_header:
            df = df.drop(0, axis = 0)
        st.dataframe(df, hide_index = True, use_container_width = True)
        st.write(f"Number of Samples: :red[{df.shape[0]}]")
        st.write(f"Number of Features: :red[{df.shape[1]}]")
        
        # st.subheader("Dataset Info")
        # with st.container(border = True):
        with st.popover("Dataset Info"):
            buffer = io.StringIO()

            # Pass the buffer to df.info()
            df.info(buf=buffer, memory_usage="deep")

            # Get the dataframe info as a string from the buffer
            df_info = buffer.getvalue()

            # Split the string into lines and remove the first line
            lines = df_info.split('\n')
            df_info = '\n'.join(lines[1:])

            # Display the dataframe info in the Streamlit app
            st.text(df_info)
        return df


from setuptools import setup, find_packages

setup(
    name="AutoMLApp",
    version="0.1.69",
    author="Shivam Nikam",
    author_email="shivam.nikam@think360.ai",
    description="An automated machine learning application designed for efficient model training, evaluation, and hyperparameter tuning.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    #url="https://github.com/YourGithub/AutoMLApp",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    install_requires=[
        'pandas==2.2.1',
        'numpy==1.26.4',
        'imbalanced-learn==0.12.0',
        'streamlit==1.32.0',
        'streamlit-extras==0.4.0',
        'streamlit-modal==0.1.2',
        'plotly==5.18.0',
        'scikit-learn==1.4.1.post1',
        'lightgbm==4.3.0',
        'xgboost==2.0.3',
        'openpyxl==3.0.10',
        'scorecardpy==0.1.9.7'  # Added to match your requirements.txt
    ],
    extras_require={
        "dev": [
            "pytest>=5.2",
            "check-manifest",
            "twine",
        ],
    },
    package_data={
        'automlapp': ['logo.png'],  # Specify relative path to package root
    },
    entry_points={
        "console_scripts": [
            "runtool=automlapp.launcher:main",
        ],
    }
)



from ..utils.streamlit_utils import st
import pandas as pd
import numpy as np
import plotly.express as px
from sklearn.metrics import confusion_matrix, accuracy_score, confusion_matrix, precision_score, recall_score, f1_score, auc, roc_curve
from sklearn.metrics import ConfusionMatrixDisplay

class ModelEvaluation:
	def __init__(_self, df, X_train, y_train, X_test, y_test):
		_self.df = df
		_self.X_train = X_train
		_self.y_train = y_train
		_self.X_test = X_test
		_self.y_test = y_test
		_self.plot_colors = px.colors.sequential.Blues[::-2]

	def evaluate_models(_self):
		if 'trained_models' in st.session_state and len(st.session_state.trained_models) != 0 and st.session_state.eval:
			target = st.session_state.target
			st.subheader("Evaluation Metrics")
			
			eval_df, eval_df_formatted = pd.DataFrame(columns=["Model", "Accuracy", "Precision", "Recall", "F1 Score"]), pd.DataFrame(columns=["Model", "Accuracy", "Precision", "Recall", "F1 Score"])
			if _self.df[target].nunique() > 2:
				avg = 'micro'
			else:
				avg = 'binary'
			for model, name in zip(st.session_state.trained_models, st.session_state.model_names):
				y_pred = model.predict(_self.X_test)
				prob = model.predict_proba(_self.X_test)[:, 1]
				fpr, tpr, thresholds = roc_curve(_self.y_test, prob)

				row_df = pd.DataFrame({
					"Model": [name],
					"Accuracy": [accuracy_score(_self.y_test, y_pred)],
					"Precision": [precision_score(_self.y_test, y_pred, average=avg)],
					"Recall": [recall_score(_self.y_test, y_pred, average=avg)],
					"F1 Score": [f1_score(_self.y_test, y_pred, average=avg)],
					"AUC": [auc(fpr, tpr)],
					"Gini": [2 * auc(fpr, tpr) - 1],
					"KS": [np.max(tpr - fpr)]
				})
				row_df_formatted = pd.DataFrame({
					"Model": [name],
					"Accuracy": [f"{accuracy_score(_self.y_test, y_pred):.2%}"],
					"Precision": [f"{precision_score(_self.y_test, y_pred, average=avg):.2%}"],
					"Recall": [f"{recall_score(_self.y_test, y_pred, average=avg):.2%}"],
					"F1 Score": [f"{f1_score(_self.y_test, y_pred, average=avg):.2%}"],
					"AUC": [f"{auc(fpr, tpr):.2%}"],
					"Gini": [f"{2 * auc(fpr, tpr) - 1:.2%}"],
					"KS": [f"{np.max(tpr - fpr):.2%}"]
				})
				eval_df = pd.concat([eval_df, row_df], ignore_index=True)
				eval_df_formatted = pd.concat([eval_df_formatted, row_df_formatted], ignore_index=True)
				st.session_state.eval_df = eval_df
			# st.table(eval_df_formatted)
			st.dataframe(eval_df_formatted, use_container_width=True)

			fig = px.bar(
				eval_df.set_index("Model"), 
				orientation='h', 
				width=980,
				color_discrete_sequence=_self.plot_colors
			)
			fig.update_layout(margin=dict(l=0, r=0, t=0, b=0), yaxis={'categoryorder':'total ascending'})
			st.write(fig)

			st.subheader("Confusion Matrix")
			cols = st.columns(len(st.session_state.trained_models))
			for i, col in enumerate(cols):
				col.write(st.session_state.model_names[i])
				y_pred = st.session_state.trained_models[i].predict(_self.X_test)
				
				# Compute the confusion matrix
				cm = confusion_matrix(_self.y_test, y_pred)
				
				# Normalize the confusion matrix by row (i.e by the number of samples in each class)
				cm_normalized = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
				
				# Create a new DataFrame for the confusion matrix
				cm_df = pd.DataFrame(cm, index=model.classes_, columns=model.classes_)
				
				# Format the DataFrame to include percentages
				for j in range(cm_df.shape[0]):
					for k in range(cm_df.shape[1]):
						cm_df.iloc[j, k] = f"{cm[j, k]} ({cm_normalized[j, k]*100:.1f}%)"
				
				# Display the confusion matrix
				col.table(cm_df)

	def evaluate_custom_models(_self):
		if 'trained_models_custom' in st.session_state and len(st.session_state.trained_models_custom) != 0:
			target = st.session_state.target
			st.subheader("Evaluation Metrics of Model Trained on Custom Hyperparameters")
			
			eval_df_custom, eval_df_custom_formatted = pd.DataFrame(columns=["Model", "Accuracy", "Precision", "Recall", "F1 Score"]), pd.DataFrame(columns=["Model", "Accuracy", "Precision", "Recall", "F1 Score"])
			if _self.df[target].nunique() > 2:
				avg = 'micro'
			else:
				avg = 'binary'
			for model, name in zip(st.session_state.trained_models_custom, st.session_state.custom_model_names):
				y_pred = model.predict(_self.X_test)
				prob = model.predict_proba(_self.X_test)[:, 1]
				fpr, tpr, thresholds = roc_curve(_self.y_test, prob)

				row_df = pd.DataFrame({
					"Model": [name],
					"Accuracy": [accuracy_score(_self.y_test, y_pred)],
					"Precision": [precision_score(_self.y_test, y_pred, average=avg)],
					"Recall": [recall_score(_self.y_test, y_pred, average=avg)],
					"F1 Score": [f1_score(_self.y_test, y_pred, average=avg)],
					"AUC": [auc(fpr, tpr)],
					"Gini": [2 * auc(fpr, tpr) - 1],
					"KS": [np.max(tpr - fpr)]
				})
				row_df_formatted = pd.DataFrame({
					"Model": [name],
					"Accuracy": [f"{accuracy_score(_self.y_test, y_pred):.2%}"],
					"Precision": [f"{precision_score(_self.y_test, y_pred, average=avg):.2%}"],
					"Recall": [f"{recall_score(_self.y_test, y_pred, average=avg):.2%}"],
					"F1 Score": [f"{f1_score(_self.y_test, y_pred, average=avg):.2%}"],
					"AUC": [f"{auc(fpr, tpr):.2%}"],
					"Gini": [f"{2 * auc(fpr, tpr) - 1:.2%}"],
					"KS": [f"{np.max(tpr - fpr):.2%}"]
				})
				eval_df_custom = pd.concat([eval_df_custom, row_df], ignore_index=True)
				eval_df_custom_formatted = pd.concat([eval_df_custom_formatted, row_df_formatted], ignore_index=True)
				st.session_state.eval_df_custom = eval_df_custom
			# st.table(eval_df_custom_formatted)
			st.dataframe(eval_df_custom_formatted, use_container_width=True)

			fig = px.bar(
				eval_df_custom.set_index("Model"), 
				orientation='h', 
				width=980,
				color_discrete_sequence=_self.plot_colors
			)
			fig.update_layout(margin=dict(l=0, r=0, t=0, b=0), yaxis={'categoryorder':'total ascending'})
			st.write(fig)

			st.subheader("Confusion Matrix")
			cols = st.columns(len(st.session_state.trained_models_custom))
			for i, col in enumerate(cols):
				col.write(st.session_state.custom_model_names[i])
				y_pred = st.session_state.trained_models_custom[i].predict(_self.X_test)
				
				# Compute the confusion matrix
				cm = confusion_matrix(_self.y_test, y_pred)
				
				# Normalize the confusion matrix by row (i.e by the number of samples in each class)
				cm_normalized = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
				
				# Create a new DataFrame for the confusion matrix
				cm_df = pd.DataFrame(cm, index=model.classes_, columns=model.classes_)
				
				# Format the DataFrame to include percentages
				for j in range(cm_df.shape[0]):
					for k in range(cm_df.shape[1]):
						cm_df.iloc[j, k] = f"{cm[j, k]} ({cm_normalized[j, k]*100:.1f}%)"
				
				# Display the confusion matrix
				col.table(cm_df)

	def evaluate_tuned_models(_self):
		if 'tuned_models' in st.session_state and len(st.session_state.tuned_models) != 0:
			target = st.session_state.target
			st.subheader("Evaluation Metrics after tuning the hyperparameters")
			
			eval_df_tuned, eval_df_tuned_formatted = pd.DataFrame(columns=["Model", "Accuracy", "Precision", "Recall", "F1 Score"]), pd.DataFrame(columns=["Model", "Accuracy", "Precision", "Recall", "F1 Score"])
			if _self.df[target].nunique() > 2:
				avg = 'micro'
			else:
				avg = 'binary'
			for model, name in zip(st.session_state.tuned_models, st.session_state.tuned_model_names):
				y_pred = model.predict(_self.X_test)
				prob = model.predict_proba(_self.X_test)[:, 1]
				fpr, tpr, thresholds = roc_curve(_self.y_test, prob)

				# Get the previous scores for the current model
				prev_scores = st.session_state.eval_df[st.session_state.eval_df['Model'] == name].iloc[0]

				# Calculate the metrics first
				accuracy = accuracy_score(_self.y_test, y_pred)
				precision = precision_score(_self.y_test, y_pred, average=avg)
				recall = recall_score(_self.y_test, y_pred, average=avg)
				f1 = f1_score(_self.y_test, y_pred, average=avg)
				auc_score = auc(fpr, tpr)
				gini = 2 * auc_score - 1
				ks = np.max(tpr - fpr)

				# Calculate the deltas
				delta_accuracy = accuracy - float(prev_scores['Accuracy'])
				delta_precision = precision - float(prev_scores['Precision'])
				delta_recall = recall - float(prev_scores['Recall'])
				delta_f1 = f1 - float(prev_scores['F1 Score'])
				delta_auc = float(auc_score - prev_scores['AUC'])
				delta_gini = gini - float(prev_scores['Gini'])
				delta_ks = ks - float(prev_scores['KS'])

				# Create a grid of columns
				cols = st.columns(7)

				# Display the metrics with deltas in the grid
				st.subheader(name)
				cols[0].metric(label = "Accuracy", value = f"{accuracy:.2%}", delta = f"{delta_accuracy:.2%}")
				cols[1].metric(label = "Precision", value = f"{precision:.2%}", delta = f"{delta_precision:.2%}")
				cols[2].metric(label = "Recall", value = f"{recall:.2%}", delta = f"{delta_recall:.2%}")
				cols[3].metric(label = "F1 Score", value = f"{f1:.2%}", delta = f"{delta_f1:.2%}")
				cols[4].metric(label = "AUC", value = f"{auc_score:.2%}", delta = f"{delta_auc:.2%}")
				cols[5].metric(label = "Gini", value = f"{gini:.2%}", delta = f"{delta_gini:.2%}")
				cols[6].metric(label = "KS", value = f"{ks:.2%}", delta = f"{delta_ks:.2%}")

				row_df = pd.DataFrame({
					"Model": [name],
					"Accuracy": [accuracy],
					"Precision": [precision],
					"Recall": [recall],
					"F1 Score": [f1],
					"AUC": [auc_score],
					"Gini": [gini],
					"KS": [ks]
				})
				row_df_formatted = pd.DataFrame({
					"Model": [name],
					"Accuracy": [f"{accuracy_score(_self.y_test, y_pred):.2%}"],
					"Precision": [f"{precision_score(_self.y_test, y_pred, average=avg):.2%}"],
					"Recall": [f"{recall_score(_self.y_test, y_pred, average=avg):.2%}"],
					"F1 Score": [f"{f1_score(_self.y_test, y_pred, average=avg):.2%}"],
					"AUC": [f"{auc(fpr, tpr):.2%}"],
					"Gini": [f"{2 * auc(fpr, tpr) - 1:.2%}"],
					"KS": [f"{np.max(tpr - fpr):.2%}"]
				})
				eval_df_tuned = pd.concat([eval_df_tuned, row_df], ignore_index=True)
				eval_df_tuned_formatted = pd.concat([eval_df_tuned_formatted, row_df_formatted], ignore_index=True)
			# st.table(eval_df_tuned_formatted)
			st.dataframe(eval_df_tuned_formatted, use_container_width=True)
			st.session_state.eval_df_tuned = eval_df_tuned_formatted

			fig = px.bar(
				eval_df_tuned.set_index("Model"), 
				orientation='h', 
				width=980,
				color_discrete_sequence=_self.plot_colors
			)
			fig.update_layout(margin=dict(l=0, r=0, t=0, b=0), yaxis={'categoryorder':'total ascending'})
			st.write(fig)

			st.subheader("Confusion Matrix")
			cols = st.columns(len(st.session_state.tuned_models))

			for i, col in enumerate(cols):
				col.write(st.session_state.tuned_model_names[i])
				y_pred = st.session_state.tuned_models[i].predict(_self.X_test)
				
				# Compute the confusion matrix
				cm = confusion_matrix(_self.y_test, y_pred)
				
				# Normalize the confusion matrix by row (i.e by the number of samples in each class)
				cm_normalized = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
				
				# Create a new DataFrame for the confusion matrix
				cm_df = pd.DataFrame(cm, index=model.classes_, columns=model.classes_)
				
				# Format the DataFrame to include percentages
				for j in range(cm_df.shape[0]):
					for k in range(cm_df.shape[1]):
						cm_df.iloc[j, k] = f"{cm[j, k]} ({cm_normalized[j, k]*100:.1f}%)"
				
				# Display the confusion matrix
				col.table(cm_df)
	
	def calculate_deciles(_self, model_type, model_name, subset):
		if model_type == "Model With Default Hyperparameters":
			model = st.session_state.trained_models_dict[str(model_name)]
		elif model_type == "Model With Tuned Hyperparameters":
			model = st.session_state.tuned_models_dict[str(model_name)]
		elif model_type == "Model With Custom Hyperparameters":
			model = st.session_state.trained_models_custom_dict[str(model_name)]
			
		if subset == "train":
			df_copy = pd.concat([_self.X_train, _self.y_train], axis = 1)
			df_copy['default'] = model.predict(_self.X_train)
			df_copy['prob'] = model.predict_proba(_self.X_train)[:, 1]
		elif subset == "test":
			df_copy = pd.concat([_self.X_test, _self.y_test], axis = 1)
			df_copy['default'] = model.predict(_self.X_test)
			df_copy['prob'] = model.predict_proba(_self.X_test)[:, 1]

		df_copy = df_copy.sort_values(by = 'prob', ascending = False)

		df_copy['decile'] = pd.qcut(df_copy['prob'], 10, labels = False)

		# Add 1 to make deciles 1 to 10
		df_copy['decile'] = df_copy['decile'] + 1

		output = df_copy.groupby('decile').agg(
			decile_score = ('decile', 'min'),
			population_count = ('decile', 'size'),
			bad_count = ('Bad_gt500_final', 'sum')
		)

		# Calculate good count
		output['good_count'] = output['population_count'] - output['bad_count']

		# Calculate response rate
		output['bad_rate'] = round((output['bad_count'] / output['population_count']) * 100, 2)

		# percentage of defaults in each decile
		total_bad = output['bad_count'].sum()
		output['bad_percentage'] = round((output['bad_count'] / total_bad) * 100, 2)

		# percentage good in each decile
		total_good = output['good_count'].sum()
		output['good_percentage'] = round((output['good_count'] / total_good) * 100, 2)

		# Sort by decile
		output.sort_values(by = 'decile', inplace = True, ascending = False)
		output.reset_index(drop = True, inplace = True)

		# Calculate cumulative good and bad count
		output['cumulative_good_count'] = output['good_count'].cumsum()
		output['cumulative_bad_count'] = output['bad_count'].cumsum()

		# Calculate cumulative good and bad percentage
		output['cumulative_bad_perc'] = output['bad_percentage'].cumsum()
		output['cumulative_good_perc'] = output['good_percentage'].cumsum()

		# Calculate cumulative sample size and cumulative default rate
		output['cumulative_population_count'] = output['population_count'].cumsum()

		# Calculate KS statistic
		output['KS'] = output['cumulative_bad_perc'] - output['cumulative_good_perc']

		result = output.loc[:,['decile_score', 'population_count', 'cumulative_population_count', 'good_count', 'bad_count', 'bad_rate', 'cumulative_good_count', 'cumulative_bad_count', 'cumulative_good_perc', 'cumulative_bad_perc', 'KS']]
		st.dataframe(result)